
open Ast

exception RunTimeError
exception Unimplemented

(* notice lookup and interp are tail-recursive (run in O(1) stack space) *)

let rec lookup (env : env) (s : variable) : valu = (* do not change this *)
  match env with
    [] -> failwith "lookup"
  | ((s2,v)::tl) -> if s2=s then v else lookup tl s

let interp (e : exp) : valu =
  let rec loop c env e =
    match e with (* you need to add more cases *)
    | S s          -> loop c env (V (lookup env s))
    | L(s1,e1)     -> loop c env (V (C(s1,e1,env)))
    | A(e1,e2)     -> loop (LeftApp(e2,env)::c) env e1
    | MP(e1,e2)    -> loop (LeftPr(e2,env)::c)  env e1
    | Fst(e1)      -> loop (SFst::c)            env e1
    | Snd(e1)      -> loop (SSnd::c)            env e1
    | If(e1,e2,e3) -> loop (SIf(e2,e3,env)::c)  env e1
    | Letcc(s,e1)  -> loop c ((s,Continuation(c))::env) e1
    | Throw(e1,e2) -> loop (LeftThrow(e2,env)::c) env e1
    | V v ->
	(match c with (* you need to add more cases *)
	  []                            -> v 
	| (LeftApp  (e2,env2))::tl      -> loop (RightApp(v)::tl) env2 e2
	| (RightApp (C(s,e2,env2)))::tl -> loop tl ((s,v)::env2) e2
	| (RightApp _)::_               -> raise RunTimeError
	| (LeftPr   (e2,env2))::tl      -> loop (RightPr(v)::tl) env2 e2
	| (RightPr  v2)::tl             -> loop tl env (V(P(v2,v)))
	| SFst :: tl -> 
	    (match v with
	      P(v1,v2) -> loop tl env (V v1)
	    | _        -> raise RunTimeError)
	| SSnd :: tl -> 
	    (match v with
	      P(v1,v2) -> loop tl env (V v2)
	    | _        -> raise RunTimeError)
	| SIf(e2,e3,env2) :: tl ->
	    (match v with
	      B true  -> loop tl env2 e2
	    | B false -> loop tl env2 e3
	    | _       -> raise RunTimeError)
	| (RightThrow (Continuation(c)))::_ -> loop c env (V v)
	| (RightThrow _) :: _               -> raise RunTimeError
	| (LeftThrow (e2,env2))::tl         -> loop (RightThrow(v)::tl) env2 e2)
  in loop [] [] e

let allow_halt e =
  Letcc("x", 
	(A(L("halt", MP(V (B true), e)),
	   L("y", (Throw(S("x"),MP(V(B false), S("y"))))))))

(*Do not change anything below here *)

(* expect one command-line argument, a file to parse *)
(* you do not need to understand this interaction with the system *)
let parse_file () =
  let argv = Sys.argv in
  let _ = 
    if Array.length argv != 2
    then (prerr_string ("usage: " ^ argv.(0) ^ " [file-to-parse]\n");
	  exit 1) in
  let ch = open_in argv.(1) in
  Parse.program Lex.lexer (Lexing.from_channel ch)

let _ =
  let prog = parse_file() in
  print_string (valu2string (interp (allow_halt prog)));
  print_newline ()
